@@ -15,17 +15,20 @@
 import copy, os, re
 from collections import OrderedDict
 import itertools, pathlib
+import hashlib
 import pickle
 from functools import lru_cache
 
 from . import environment
 from . import dependencies
 from . import mlog
-from .mesonlib import File, MesonException, listify, extract_as_list, OrderedSet
-from .mesonlib import typeslistify, stringlistify, classify_unity_sources
-from .mesonlib import get_filenames_templates_dict, substitute_values
-from .mesonlib import for_windows, for_darwin, for_cygwin, for_android, has_path_sep
-from .compilers import is_object, clink_langs, sort_clink, lang_suffixes
+from .mesonlib import (
+    File, MesonException, listify, extract_as_list, OrderedSet,
+    typeslistify, stringlistify, classify_unity_sources,
+    get_filenames_templates_dict, substitute_values,
+    for_windows, for_darwin, for_cygwin, for_android, has_path_sep
+)
+from .compilers import is_object, clink_langs, sort_clink, lang_suffixes, get_macos_dylib_install_name
 from .interpreterbase import FeatureNew
 
 pch_kwargs = set(['c_pch', 'cpp_pch'])
@@ -37,6 +40,7 @@ lang_arg_kwargs = set([
     'd_import_dirs',
     'd_unittest',
     'd_module_versions',
+    'd_debug',
     'fortran_args',
     'java_args',
     'objc_args',
@@ -83,12 +87,16 @@ known_build_target_kwargs = (
     rust_kwargs |
     cs_kwargs)
 
-known_exe_kwargs = known_build_target_kwargs | {'implib', 'export_dynamic'}
+known_exe_kwargs = known_build_target_kwargs | {'implib', 'export_dynamic', 'pie'}
 known_shlib_kwargs = known_build_target_kwargs | {'version', 'soversion', 'vs_module_defs', 'darwin_versions'}
 known_shmod_kwargs = known_build_target_kwargs
 known_stlib_kwargs = known_build_target_kwargs | {'pic'}
 known_jar_kwargs = known_exe_kwargs | {'main_class'}
 
+@lru_cache(maxsize=None)
+def get_target_macos_dylib_install_name(ld):
+    return get_macos_dylib_install_name(ld.prefix, ld.name, ld.suffix, ld.soversion)
+
 class InvalidArguments(MesonException):
     pass
 
@@ -130,9 +138,23 @@ class Build:
         self.dep_manifest = {}
         self.cross_stdlibs = {}
         self.test_setups = {}
+        self.test_setup_default_name = None
         self.find_overrides = {}
         self.searched_programs = set() # The list of all programs that have been searched for.
 
+    def copy(self):
+        other = Build(self.environment)
+        for k, v in self.__dict__.items():
+            if isinstance(v, (list, dict, set, OrderedDict)):
+                other.__dict__[k] = v.copy()
+            else:
+                other.__dict__[k] = v
+        return other
+
+    def merge(self, other):
+        for k, v in other.__dict__.items():
+            self.__dict__[k] = v
+
     def add_compiler(self, compiler):
         if self.static_linker is None and compiler.needs_static_linker():
             self.static_linker = self.environment.detect_static_linker(compiler)
@@ -324,25 +346,58 @@ a hard error in the future.''' % name)
         self.build_always_stale = False
         self.option_overrides = {}
 
+    def get_install_dir(self, environment):
+        # Find the installation directory.
+        default_install_dir = self.get_default_install_dir(environment)
+        outdirs = self.get_custom_install_dir()
+        if outdirs[0] is not None and outdirs[0] != default_install_dir and outdirs[0] is not True:
+            # Either the value is set to a non-default value, or is set to
+            # False (which means we want this specific output out of many
+            # outputs to not be installed).
+            custom_install_dir = True
+        else:
+            custom_install_dir = False
+            outdirs[0] = default_install_dir
+        return outdirs, custom_install_dir
+
     def get_basename(self):
         return self.name
 
     def get_subdir(self):
         return self.subdir
 
-    def get_id(self):
+    @staticmethod
+    def _get_id_hash(target_id):
+        # We don't really need cryptographic security here.
+        # Small-digest hash function with unlikely collision is good enough.
+        h = hashlib.sha256()
+        h.update(target_id.encode(encoding='utf-8', errors='replace'))
+        # This ID should be case-insensitive and should work in Visual Studio,
+        # e.g. it should not start with leading '-'.
+        return h.hexdigest()[:7]
+
+    @staticmethod
+    def construct_id_from_path(subdir, name, type_suffix):
+        """Construct target ID from subdir, name and type suffix.
+
+        This helper function is made public mostly for tests."""
         # This ID must also be a valid file name on all OSs.
         # It should also avoid shell metacharacters for obvious
         # reasons. '@' is not used as often as '_' in source code names.
         # In case of collisions consider using checksums.
         # FIXME replace with assert when slash in names is prohibited
-        name_part = self.name.replace('/', '@').replace('\\', '@')
-        assert not has_path_sep(self.type_suffix())
-        myid = name_part + self.type_suffix()
-        if self.subdir:
-            subdir_part = self.subdir.replace('/', '@').replace('\\', '@')
-            myid = subdir_part + '@@' + myid
-        return myid
+        name_part = name.replace('/', '@').replace('\\', '@')
+        assert not has_path_sep(type_suffix)
+        my_id = name_part + type_suffix
+        if subdir:
+            subdir_part = Target._get_id_hash(subdir)
+            # preserve myid for better debuggability
+            return subdir_part + '@@' + my_id
+        return my_id
+
+    def get_id(self):
+        return self.construct_id_from_path(
+            self.subdir, self.name, self.type_suffix())
 
     def process_kwargs(self, kwargs):
         if 'build_by_default' in kwargs:
@@ -395,6 +450,8 @@ class BuildTarget(Target):
         self.generated = []
         self.extra_files = []
         self.d_features = {}
+        self.pic = False
+        self.pie = False
         # Sources can be:
         # 1. Pre-existing source files in the source tree
         # 2. Pre-existing sources generated by configure_file in the build tree
@@ -422,7 +479,7 @@ class BuildTarget(Target):
         return repr_str.format(self.__class__.__name__, self.get_id(), self.filename)
 
     def validate_cross_install(self, environment):
-        if environment.is_cross_build() and not self.is_cross and self.install:
+        if environment.is_cross_build() and not self.is_cross and self.need_install:
             raise InvalidArguments('Tried to install a natively built target in a cross build.')
 
     def check_unknown_kwargs(self, kwargs):
@@ -678,6 +735,20 @@ class BuildTarget(Target):
             result += i.get_all_link_deps()
         return result
 
+    def get_link_deps_mapping(self, prefix, environment):
+        return self.get_transitive_link_deps_mapping(prefix, environment)
+
+    @lru_cache(maxsize=None)
+    def get_transitive_link_deps_mapping(self, prefix, environment):
+        result = {}
+        for i in self.link_targets:
+            mapping = i.get_link_deps_mapping(prefix, environment)
+            #we are merging two dictionaries, while keeping the earlier one dominant
+            result_tmp = mapping.copy()
+            result_tmp.update(result)
+            result = result_tmp
+        return result
+
     @lru_cache(maxsize=None)
     def get_link_dep_subdirs(self):
         result = OrderedSet()
@@ -686,6 +757,9 @@ class BuildTarget(Target):
             result.update(i.get_link_dep_subdirs())
         return result
 
+    def get_default_install_dir(self, environment):
+        return environment.get_libdir()
+
     def get_custom_install_dir(self):
         return self.install_dir
 
@@ -737,9 +811,12 @@ just like those detected with the dependency() function.''')
         dfeature_unittest = kwargs.get('d_unittest', False)
         if dfeature_unittest:
             dfeatures['unittest'] = dfeature_unittest
-        dfeature_versions = kwargs.get('d_module_versions', None)
+        dfeature_versions = kwargs.get('d_module_versions', [])
         if dfeature_versions:
             dfeatures['versions'] = dfeature_versions
+        dfeature_debug = kwargs.get('d_debug', [])
+        if dfeature_debug:
+            dfeatures['debug'] = dfeature_debug
         if 'd_import_dirs' in kwargs:
             dfeature_import_dirs = extract_as_list(kwargs, 'd_import_dirs', unholder=True)
             for d in dfeature_import_dirs:
@@ -830,13 +907,14 @@ This will become a hard error in a future Meson release.''')
             # since library loading is done differently)
             if for_darwin(self.is_cross, self.environment) or for_windows(self.is_cross, self.environment):
                 self.pic = True
-            elif '-fPIC' in clist + cpplist:
-                mlog.warning("Use the 'pic' kwarg instead of passing -fPIC manually to static library {!r}".format(self.name))
-                self.pic = True
             else:
-                self.pic = kwargs.get('pic', False)
-                if not isinstance(self.pic, bool):
-                    raise InvalidArguments('Argument pic to static library {!r} must be boolean'.format(self.name))
+                self.pic = self._extract_pic_pie(kwargs, 'pic')
+        if isinstance(self, Executable):
+            # Executables must be PIE on Android
+            if for_android(self.is_cross, self.environment):
+                self.pie = True
+            else:
+                self.pie = self._extract_pic_pie(kwargs, 'pie')
         self.implicit_include_directories = kwargs.get('implicit_include_directories', True)
         if not isinstance(self.implicit_include_directories, bool):
             raise InvalidArguments('Implicit_include_directories must be a boolean.')
@@ -849,6 +927,18 @@ This will become a hard error in a future Meson release.''')
                 raise InvalidArguments('GNU symbol visibility arg %s not one of: %s',
                                        self.symbol_visibility, ', '.join(permitted))
 
+    def _extract_pic_pie(self, kwargs, arg):
+        # Check if we have -fPIC, -fpic, -fPIE, or -fpie in cflags
+        all_flags = self.extra_args['c'] + self.extra_args['cpp']
+        if '-f' + arg.lower() in all_flags or '-f' + arg.upper() in all_flags:
+            mlog.warning("Use the '{}' kwarg instead of passing '{}' manually to {!r}".format(arg, '-f' + arg, self.name))
+            return True
+
+        val = kwargs.get(arg, False)
+        if not isinstance(val, bool):
+            raise InvalidArguments('Argument {} to {!r} must be boolean'.format(arg, self.name))
+        return val
+
     def get_filename(self):
         return self.filename
 
@@ -1101,7 +1191,7 @@ You probably should put it in link_with instead.''')
         '''
         linker, _ = self.get_clink_dynamic_linker_and_stdlibs()
         # Mixing many languages with MSVC is not supported yet so ignore stdlibs.
-        if linker and linker.get_id() in ['msvc', 'llvm', 'dmd']:
+        if linker and linker.get_id() in ['msvc', 'clang-cl', 'llvm', 'dmd']:
             return True
         return False
 
@@ -1268,6 +1358,8 @@ class Executable(BuildTarget):
     known_kwargs = known_exe_kwargs
 
     def __init__(self, name, subdir, subproject, is_cross, sources, objects, environment, kwargs):
+        if 'pie' not in kwargs and 'b_pie' in environment.coredata.base_options:
+            kwargs['pie'] = environment.coredata.base_options['b_pie'].value
         super().__init__(name, subdir, subproject, is_cross, sources, objects, environment, kwargs)
         # Unless overridden, executables have no suffix or prefix. Except on
         # Windows and with C#/Mono executables where the suffix is 'exe'
@@ -1278,12 +1370,14 @@ class Executable(BuildTarget):
             if (for_windows(is_cross, environment) or
                     for_cygwin(is_cross, environment) or 'cs' in self.compilers):
                 self.suffix = 'exe'
+            elif ('c' in self.compilers and self.compilers['c'].get_id().startswith('arm') or
+                  'cpp' in self.compilers and self.compilers['cpp'].get_id().startswith('arm')):
+                self.suffix = 'axf'
+            elif ('c' in self.compilers and self.compilers['c'].get_id().startswith('ccrx') or
+                  'cpp' in self.compilers and self.compilers['cpp'].get_id().startswith('ccrx')):
+                self.suffix = 'abs'
             else:
-                if ('c' in self.compilers and self.compilers['c'].get_id().startswith('arm') or
-                   'cpp' in self.compilers and self.compilers['cpp'].get_id().startswith('arm')):
-                    self.suffix = 'axf'
-                else:
-                    self.suffix = ''
+                self.suffix = ''
         self.filename = self.name
         if self.suffix:
             self.filename += '.' + self.suffix
@@ -1323,6 +1417,9 @@ class Executable(BuildTarget):
         # Only linkwithable if using export_dynamic
         self.is_linkwithable = self.export_dynamic
 
+    def get_default_install_dir(self, environment):
+        return environment.get_bindir()
+
     def description(self):
         '''Human friendly description of the executable'''
         return self.name
@@ -1384,6 +1481,12 @@ class StaticLibrary(BuildTarget):
         self.filename = self.prefix + self.name + '.' + self.suffix
         self.outputs = [self.filename]
 
+    def get_link_deps_mapping(self, prefix, environment):
+        return {}
+
+    def get_default_install_dir(self, environment):
+        return environment.get_static_lib_dir()
+
     def type_suffix(self):
         return "@sta"
 
@@ -1430,6 +1533,21 @@ class SharedLibrary(BuildTarget):
         self.basic_filename_tpl = '{0.prefix}{0.name}.{0.suffix}'
         self.determine_filenames(is_cross, environment)
 
+    def get_link_deps_mapping(self, prefix, environment):
+        result = {}
+        mappings = self.get_transitive_link_deps_mapping(prefix, environment)
+        old = get_target_macos_dylib_install_name(self)
+        if old not in mappings:
+            fname = self.get_filename()
+            outdirs, _ = self.get_install_dir(self.environment)
+            new = os.path.join(prefix, outdirs[0], fname)
+            result.update({old: new})
+        mappings.update(result)
+        return mappings
+
+    def get_default_install_dir(self, environment):
+        return environment.get_shared_lib_dir()
+
     def determine_filenames(self, is_cross, env):
         """
         See https://github.com/mesonbuild/meson/pull/417 for details.
@@ -1454,13 +1572,9 @@ class SharedLibrary(BuildTarget):
         prefix = ''
         suffix = ''
         self.filename_tpl = self.basic_filename_tpl
-        # If the user already provided the prefix and suffix to us, we don't
-        # need to do any filename suffix/prefix detection.
         # NOTE: manual prefix/suffix override is currently only tested for C/C++
-        if self.prefix is not None and self.suffix is not None:
-            pass
         # C# and Mono
-        elif 'cs' in self.compilers:
+        if 'cs' in self.compilers:
             prefix = ''
             suffix = 'dll'
             self.filename_tpl = '{0.prefix}{0.name}.{0.suffix}'
@@ -1469,8 +1583,8 @@ class SharedLibrary(BuildTarget):
         # For all other targets/platforms import_filename stays None
         elif for_windows(is_cross, env):
             suffix = 'dll'
-            self.vs_import_filename = '{0}.lib'.format(self.name)
-            self.gcc_import_filename = 'lib{0}.dll.a'.format(self.name)
+            self.vs_import_filename = '{0}{1}.lib'.format(self.prefix if self.prefix is not None else '', self.name)
+            self.gcc_import_filename = '{0}{1}.dll.a'.format(self.prefix if self.prefix is not None else 'lib', self.name)
             if self.get_using_msvc():
                 # Shared library is of the form foo.dll
                 prefix = ''
@@ -1489,7 +1603,7 @@ class SharedLibrary(BuildTarget):
                 self.filename_tpl = '{0.prefix}{0.name}.{0.suffix}'
         elif for_cygwin(is_cross, env):
             suffix = 'dll'
-            self.gcc_import_filename = 'lib{0}.dll.a'.format(self.name)
+            self.gcc_import_filename = '{0}{1}.dll.a'.format(self.prefix if self.prefix is not None else 'lib', self.name)
             # Shared library is of the form cygfoo.dll
             # (ld --dll-search-prefix=cyg is the default)
             prefix = 'cyg'
@@ -1701,6 +1815,10 @@ class SharedModule(SharedLibrary):
             raise MesonException('Shared modules must not specify the soversion kwarg.')
         super().__init__(name, subdir, subproject, is_cross, sources, objects, environment, kwargs)
 
+    def get_default_install_dir(self, environment):
+        return environment.get_shared_module_dir()
+
+
 class CustomTarget(Target):
     known_kwargs = set([
         'input',
@@ -1738,6 +1856,9 @@ class CustomTarget(Target):
             mlog.warning('Unknown keyword arguments in target %s: %s' %
                          (self.name, ', '.join(unknowns)))
 
+    def get_default_install_dir(self, environment):
+        return None
+
     def __lt__(self, other):
         return self.get_id() < other.get_id()
 
@@ -2022,7 +2143,10 @@ class Jar(BuildTarget):
 
     def get_classpath_args(self):
         cp_paths = [os.path.join(l.get_subdir(), l.get_filename()) for l in self.link_targets]
-        return ['-cp', os.pathsep.join(cp_paths)]
+        cp_string = os.pathsep.join(cp_paths)
+        if cp_string:
+            return ['-cp', os.pathsep.join(cp_paths)]
+        return []
 
 class CustomTargetIndex:
 
