@@ -17,13 +17,15 @@ import pickle, os, uuid, shlex
 import sys
 from pathlib import PurePath
 from collections import OrderedDict
-from .mesonlib import MesonException
-from .mesonlib import default_libdir, default_libexecdir, default_prefix
+from .mesonlib import (
+    MesonException, default_libdir, default_libexecdir, default_prefix
+)
 from .wrap import WrapMode
 import ast
 import argparse
+import configparser
 
-version = '0.48.2'
+version = '0.49.0'
 backendlist = ['ninja', 'vs', 'vs2010', 'vs2015', 'vs2017', 'xcode']
 
 default_yielding = False
@@ -182,6 +184,7 @@ class UserArrayOption(UserOption):
                     ', '.join(bad), ', '.join(self.choices)))
         return newvalue
 
+
 class UserFeatureOption(UserComboOption):
     static_choices = ['enabled', 'disabled', 'auto']
 
@@ -197,6 +200,72 @@ class UserFeatureOption(UserComboOption):
     def is_auto(self):
         return self.value == 'auto'
 
+
+def load_configs(filenames):
+    """Load native files."""
+    def gen():
+        for f in filenames:
+            f = os.path.expanduser(os.path.expandvars(f))
+            if os.path.exists(f):
+                yield f
+                continue
+            elif sys.platform != 'win32':
+                f = os.path.basename(f)
+                paths = [
+                    os.environ.get('XDG_DATA_HOME', os.path.expanduser('~/.local/share')),
+                ] + os.environ.get('XDG_DATA_DIRS', '/usr/local/share:/usr/share').split(':')
+                for path in paths:
+                    path_to_try = os.path.join(path, 'meson', 'native', f)
+                    if os.path.isfile(path_to_try):
+                        yield path_to_try
+                        break
+                else:
+                    raise MesonException('Cannot find specified native file: ' + f)
+                continue
+
+            raise MesonException('Cannot find specified native file: ' + f)
+
+    config = configparser.SafeConfigParser()
+    config.read(gen())
+    return config
+
+
+def _get_section(config, section):
+    if config.has_section(section):
+        final = {}
+        for k, v in config.items(section):
+            # Windows paths...
+            v = v.replace('\\', '\\\\')
+            try:
+                final[k] = ast.literal_eval(v)
+            except SyntaxError:
+                raise MesonException(
+                    'Malformed value in native file variable: {}'.format(v))
+        return final
+    return {}
+
+
+class ConfigData:
+
+    """Contains configuration information provided by the user for the build."""
+
+    def __init__(self, config=None):
+        if config:
+            self.binaries = _get_section(config, 'binaries')
+            # global is a keyword and globals is a builtin, rather than mangle it,
+            # use a similar word
+            self.universal = _get_section(config, 'globals')
+            self.subprojects = {s: _get_section(config, s) for s in config.sections()
+                                if s not in {'binaries', 'globals'}}
+        else:
+            self.binaries = {}
+            self.universal = {}
+            self.subprojects = {}
+
+    def get_binaries(self, name):
+        return self.binaries.get(name, None)
+
+
 # This class contains all data that must persist over multiple
 # invocations of Meson. It is roughly the same thing as
 # cmakecache.
@@ -223,12 +292,20 @@ class CoreData:
         self.base_options = {}
         self.external_preprocess_args = {} # CPPFLAGS only
         self.cross_file = self.__load_cross_file(options.cross_file)
-        self.wrap_mode = options.wrap_mode if options.wrap_mode is not None else WrapMode.default
         self.compilers = OrderedDict()
         self.cross_compilers = OrderedDict()
         self.deps = OrderedDict()
         # Only to print a warning if it changes between Meson invocations.
         self.pkgconf_envvar = os.environ.get('PKG_CONFIG_PATH', '')
+        self.config_files = self.__load_config_files(options.native_file)
+
+    @staticmethod
+    def __load_config_files(filenames):
+        if not filenames:
+            return []
+        filenames = [os.path.abspath(os.path.expanduser(os.path.expanduser(f)))
+                     for f in filenames]
+        return filenames
 
     @staticmethod
     def __load_cross_file(filename):
@@ -338,7 +415,10 @@ class CoreData:
 
     def get_builtin_option(self, optname):
         if optname in self.builtins:
-            return self.builtins[optname].value
+            v = self.builtins[optname]
+            if optname == 'wrap_mode':
+                return WrapMode.from_string(v.value)
+            return v.value
         raise RuntimeError('Tried to get unknown builtin option %s.' % optname)
 
     def set_builtin_option(self, optname, value):
@@ -457,6 +537,51 @@ class CoreData:
             sub = 'In subproject {}: '.format(subproject) if subproject else ''
             mlog.warning('{}Unknown options: "{}"'.format(sub, unknown_options))
 
+class CmdLineFileParser(configparser.ConfigParser):
+    def __init__(self):
+        # We don't want ':' as key delimiter, otherwise it would break when
+        # storing subproject options like "subproject:option=value"
+        super().__init__(delimiters=['='])
+
+def get_cmd_line_file(build_dir):
+    return os.path.join(build_dir, 'meson-private', 'cmd_line.txt')
+
+def read_cmd_line_file(build_dir, options):
+    filename = get_cmd_line_file(build_dir)
+    config = CmdLineFileParser()
+    config.read(filename)
+
+    # Do a copy because config is not really a dict. options.cmd_line_options
+    # overrides values from the file.
+    d = dict(config['options'])
+    d.update(options.cmd_line_options)
+    options.cmd_line_options = d
+
+    properties = config['properties']
+    if options.cross_file is None:
+        options.cross_file = properties.get('cross_file', None)
+
+def write_cmd_line_file(build_dir, options):
+    filename = get_cmd_line_file(build_dir)
+    config = CmdLineFileParser()
+
+    properties = {}
+    if options.cross_file is not None:
+        properties['cross_file'] = options.cross_file
+
+    config['options'] = options.cmd_line_options
+    config['properties'] = properties
+    with open(filename, 'w') as f:
+        config.write(f)
+
+def update_cmd_line_file(build_dir, options):
+    filename = get_cmd_line_file(build_dir)
+    config = CmdLineFileParser()
+    config.read(filename)
+    config['options'].update(options.cmd_line_options)
+    with open(filename, 'w') as f:
+        config.write(f)
+
 def load(build_dir):
     filename = os.path.join(build_dir, 'meson-private', 'coredata.dat')
     load_fail_msg = 'Coredata file {!r} is corrupted. Try with a fresh build tree.'.format(filename)
@@ -468,7 +593,8 @@ def load(build_dir):
     if not isinstance(obj, CoreData):
         raise MesonException(load_fail_msg)
     if obj.version != version:
-        raise MesonException('Build directory has been generated with Meson version %s, which is incompatible with current version %s.\nPlease delete this build directory AND create a new one.' %
+        raise MesonException('Build directory has been generated with Meson version %s, '
+                             'which is incompatible with current version %s.\n' %
                              (obj.version, version))
     return obj
 
@@ -616,7 +742,11 @@ builtin_options = {
     'install_umask':   [UserUmaskOption, 'Default umask to apply on permissions of installed files', '022'],
     'auto_features':   [UserFeatureOption, "Override value of all 'auto' features", 'auto'],
     'optimization':    [UserComboOption, 'Optimization level', ['0', 'g', '1', '2', '3', 's'], '0'],
-    'debug':           [UserBooleanOption, 'Debug', True]
+    'debug':           [UserBooleanOption, 'Debug', True],
+    'wrap_mode':       [UserComboOption, 'Wrap mode', ['default',
+                                                       'nofallback',
+                                                       'nodownload',
+                                                       'forcefallback'], 'default'],
 }
 
 # Special prefix-dependent defaults for installation directories that reside in
