@@ -132,7 +132,6 @@ class BoostDependency(ExternalDependency):
                 self.incdir = self.detect_nix_incdir()
 
         if self.check_invalid_modules():
-            self.log_fail()
             return
 
         mlog.debug('Boost library root dir is', mlog.bold(self.boost_root))
@@ -146,12 +145,6 @@ class BoostDependency(ExternalDependency):
             self.detect_lib_modules()
             mlog.debug('Boost library directory is', mlog.bold(self.libdir))
 
-        # 3. Report success or failure
-        if self.is_found:
-            self.log_success()
-        else:
-            self.log_fail()
-
     def check_invalid_modules(self):
         invalid_modules = [c for c in self.requested_modules if 'boost_' + c not in BOOST_LIBS]
 
@@ -172,17 +165,14 @@ class BoostDependency(ExternalDependency):
         else:
             return False
 
-    def log_fail(self):
+    def log_details(self):
         module_str = ', '.join(self.requested_modules)
-        mlog.log("Dependency Boost (%s) found:" % module_str, mlog.red('NO'))
+        return module_str
 
-    def log_success(self):
-        module_str = ', '.join(self.requested_modules)
+    def log_info(self):
         if self.boost_root:
-            info = self.version + ', ' + self.boost_root
-        else:
-            info = self.version
-        mlog.log('Dependency Boost (%s) found:' % module_str, mlog.green('YES'), info)
+            return self.boost_root
+        return ''
 
     def detect_nix_roots(self):
         return [os.path.abspath(os.path.join(x, '..'))
@@ -298,7 +288,7 @@ class BoostDependency(ExternalDependency):
         tag = None
         compiler = self.env.detect_cpp_compiler(self.want_cross)
         if mesonlib.for_windows(self.want_cross, self.env):
-            if compiler.get_id() == 'msvc':
+            if compiler.get_id() in ['msvc', 'clang-cl']:
                 comp_ts_version = compiler.get_toolset_version()
                 compiler_ts = comp_ts_version.split('.')
                 # FIXME - what about other compilers?
@@ -330,7 +320,7 @@ class BoostDependency(ExternalDependency):
 
     def arch_tag(self):
         # currently only applies to windows msvc installed binaries
-        if self.env.detect_cpp_compiler(self.want_cross).get_id() != 'msvc':
+        if self.env.detect_cpp_compiler(self.want_cross).get_id() not in ['msvc', 'clang-cl']:
             return ''
         # pre-compiled binaries only added arch tag for versions > 1.64
         if float(self.version) < 1.65:
@@ -453,7 +443,7 @@ class BoostDependency(ExternalDependency):
         if self.libdir:
             libdirs = [self.libdir]
         elif self.boost_root is None:
-            libdirs = mesonlib.get_library_dirs(self.env)
+            libdirs = mesonlib.get_library_dirs()
         else:
             libdirs = [os.path.join(self.boost_root, 'lib')]
         for libdir in libdirs:
