@@ -5,5 +5,5 @@
 import sys
 
 fc = open('/etc/apt/sources.list').read()
-if 'artful' not in fc and 'bionic' not in fc:
+if 'artful' not in fc and 'bionic' not in fc and 'cosmic' not in fc:
     sys.exit(1)
